@@ -17,6 +17,7 @@ from . import xmlwriter
 from .utility import xl_rowcol_to_cell
 from .utility import xl_cell_to_rowcol
 from .utility import xl_col_to_name
+from .utility import xl_range
 
 
 ###############################################################################
@@ -225,10 +226,13 @@ class Worksheet(xmlwriter.XMLwriter):
         self.vml_shape_id = 1024
         self.buttons_array = []
 
-        self.autofilter = ''
+        self.autofilter_area = ''
+        self.autofilter_ref = None
+        self.filter_range = []
         self.filter_on = 0
         self.filter_range = []
         self.filter_cols = {}
+        self.filter_type = {}
 
         self.col_sizes = {}
         self.row_sizes = {}
@@ -299,18 +303,20 @@ class Worksheet(xmlwriter.XMLwriter):
         if token is None:
             token = ''
 
-        # First check if the token to write is a number.
+        # Check for a datetime object.
+        if isinstance(token, datetime.datetime):
+            return self.write_datetime(row, col, *args)
+
+        # Then check if the token to write is a number.
         try:
             float(token)
             return self.write_number(row, col, *args)
-        except:
+        except ValueError:
             # Not a number. Continue to the checks below.
             pass
 
         # Map the data to the appropriate write_*() method.
-        if isinstance(token, datetime.datetime):
-            return self.write_datetime(row, col, *args)
-        elif token == '':
+        if token == '':
             return self.write_blank(row, col, *args)
         elif token.startswith('='):
             return self.write_formula(row, col, *args)
@@ -385,6 +391,9 @@ class Worksheet(xmlwriter.XMLwriter):
             -1: Row or column is out of worksheet bounds.
 
         """
+        # TODO catch and re-raise exception if token isn't a number.
+        number = float(number)
+
         # Check that row and col are valid and store max and min values.
         if self._check_dimensions(row, col):
             return -1
@@ -720,6 +729,52 @@ class Worksheet(xmlwriter.XMLwriter):
 
         return str_error
 
+    @convert_cell_args
+    def write_row(self, row, col, data, cell_format=None):
+        """
+        Write a row of data starting from (row, col).
+
+        Args:
+            row:    The cell row (zero indexed).
+            col:    The cell column (zero indexed).
+            data:   A list of tokens to be written with write().
+            format: An optional cell Format object.
+        Returns:
+            0:  Success.
+            other: Return value of write() method.
+
+        """
+        for token in (data):
+            error = self.write(row, col, token, cell_format)
+            if error:
+                return error
+            col += 1
+
+        return 0
+
+    @convert_cell_args
+    def write_column(self, row, col, data, cell_format=None):
+        """
+        Write a column of data starting from (row, col).
+
+        Args:
+            row:    The cell row (zero indexed).
+            col:    The cell column (zero indexed).
+            data:   A list of tokens to be written with write().
+            format: An optional cell Format object.
+        Returns:
+            0:  Success.
+            other: Return value of write() method.
+
+        """
+        for token in (data):
+            error = self.write(row, col, token, cell_format)
+            if error:
+                return error
+            row += 1
+
+        return 0
+
     def activate(self):
         """
         Set this worksheet as the active worksheet, i.e. the worksheet that is
@@ -756,7 +811,7 @@ class Worksheet(xmlwriter.XMLwriter):
         self.hidden = 0
 
     @convert_column_args
-    def set_column(self, firstcol, lastcol, width, cell_format=None,
+    def set_column(self, firstcol, lastcol, width=None, cell_format=None,
                    options={}):
         """
         Set the width, and other properties of a single column or a
@@ -765,7 +820,7 @@ class Worksheet(xmlwriter.XMLwriter):
         Args:
             firstcol:    First column (zero-indexed).
             lastcol:     Last column (zero-indexed). Can be same as firstcol.
-            width:       Column width.
+            width:       Column width. (optional).
             cell_format: Column cell_format. (optional).
             options:     Dict of options such as hidden and level.
 
@@ -827,14 +882,14 @@ class Worksheet(xmlwriter.XMLwriter):
 
         return 0
 
-    def set_row(self, row, height, cell_format=None, options={}):
+    def set_row(self, row, height=None, cell_format=None, options={}):
         """
         Set the width, and other properties of a row.
         range of columns.
 
         Args:
             row:         Row number (zero-indexed).
-            height:      Row width.
+            height:      Row width. (optional).
             cell_format: Row cell_format. (optional).
             options:     Dict of options such as hidden, level and collapsed.
 
@@ -935,6 +990,135 @@ class Worksheet(xmlwriter.XMLwriter):
                     continue
                 self.write_blank(row, col, '', cell_format)
 
+    @convert_range_args
+    def autofilter(self, first_row, first_col, last_row, last_col):
+        """
+        Set the autofilter area in the worksheet.
+
+        Args:
+            first_row:    The first row of the cell range. (zero indexed).
+            first_col:    The first column of the cell range.
+            last_row:     The last row of the cell range. (zero indexed).
+            last_col:     The last column of the cell range.
+
+        Returns:
+             Nothing.
+
+        """
+        # Reverse max and min values if necessary.
+        if last_row < first_row:
+            (first_row, last_row) = (last_row, first_row)
+        if last_col < first_col:
+            (first_col, last_col) = (last_col, first_col)
+
+        # Build up the print area range "Sheet1!$A$1:$C$13".
+        area = self._convert_name_area(first_row, first_col,
+                                       last_row, last_col)
+        ref = xl_range(first_row, first_col, last_row, last_col)
+
+        self.autofilter_area = area
+        self.autofilter_ref = ref
+        self.filter_range = [first_col, last_col]
+
+    def filter_column(self, col, expression):
+        """
+        Set the column filter criteria.
+
+        Args:
+            col:         Filter column (zero-indexed).
+            expression:  Filter criteria.
+
+        Returns:
+             Nothing.
+
+        """
+        if not self.autofilter_area:
+            warn("Must call autofilter() before filter_column()")
+            return
+
+        # Check for a column reference in A1 notation and substitute.
+        try:
+            int(col)
+        except ValueError:
+            # Convert col ref to a cell ref and then to a col number.
+            col_letter = col
+            (_, col) = xl_cell_to_rowcol(col + '1')
+
+            if col >= self.xls_colmax:
+                warn("Invalid column '%d'" % col_letter)
+                return
+
+        (col_first, col_last) = self.filter_range
+
+        # Reject column if it is outside filter range.
+        if col < col_first or col > col_last:
+            warn("Column '%d' outside autofilter() column range (%d, %d)"
+                 % (col, col_first, col_last))
+            return
+
+        tokens = self._extract_filter_tokens(expression)
+
+        if not (len(tokens) == 3 or len(tokens) == 7):
+            warn("Incorrect number of tokens in expression '%s'" % expression)
+
+        tokens = self._parse_filter_expression(expression, tokens)
+
+        # Excel handles single or double custom filters as default filters.
+        #  We need to check for them and handle them accordingly.
+        if len(tokens) == 2 and tokens[0] == 2:
+            # Single equality.
+            self.filter_column_list(col, [tokens[1]])
+        elif (len(tokens) == 5 and tokens[0] == 2 and tokens[2] == 1
+              and tokens[3] == 2):
+            # Double equality with "or" operator.
+            self.filter_column_list(col, [tokens[1], tokens[4]])
+        else:
+            # Non default custom filter.
+            self.filter_cols[col] = tokens
+            self.filter_type[col] = 0
+
+        self.filter_on = 1
+
+    def filter_column_list(self, col, filters):
+        """
+        Set the column filter criteria in Excel 2007 list style.
+
+        Args:
+            col:      Filter column (zero-indexed).
+            filters:  List of filter criteria to match.
+
+        Returns:
+             Nothing.
+
+        """
+        if not self.autofilter_area:
+            warn("Must call autofilter() before filter_column()")
+            return
+
+        # Check for a column reference in A1 notation and substitute.
+        try:
+            int(col)
+        except ValueError:
+            # Convert col ref to a cell ref and then to a col number.
+            col_letter = col
+            (_, col) = xl_cell_to_rowcol(col + '1')
+
+            if col >= self.xls_colmax:
+                warn("Invalid column '%d'" % col_letter)
+                return
+
+        (col_first, col_last) = self.filter_range
+
+        # Reject column if it is outside filter range.
+        if col < col_first or col > col_last:
+            warn("Column '%d' outside autofilter() column range "
+                 "(%d,%d)" % (col, col_first, col_last))
+            return
+
+        self.filter_cols[col] = filters
+        self.filter_type[col] = 1
+        self.filter_on = 1
+
     ###########################################################################
     #
     # Public API. Page Setup methods.
@@ -1354,7 +1538,7 @@ class Worksheet(xmlwriter.XMLwriter):
         # self._write_phonetic_pr()
 
         # Write the autoFilter element.
-        # self._write_auto_filter()
+        self._write_auto_filter()
 
         # Write the mergeCells element.
         self._write_merge_cells()
@@ -1579,6 +1763,149 @@ class Worksheet(xmlwriter.XMLwriter):
 
         return breaks_list
 
+    def _extract_filter_tokens(self, expression):
+        # Extract the tokens from the filter expression. The tokens are mainly
+        # non-whitespace groups. The only tricky part is to extract string
+        # tokens that contain whitespace and/or quoted double quotes (Excel's
+        # escaped quotes).
+        #
+        # Examples: 'x <  2000'
+        #           'x >  2000 and x <  5000'
+        #           'x = "foo"'
+        #           'x = "foo bar"'
+        #           'x = "foo "" bar"'
+        #
+        if not expression:
+            return []
+
+        token_re = re.compile(r'"(?:[^"]|"")*"|\S+')
+        tokens = token_re.findall(expression)
+
+        new_tokens = []
+        # Remove single leading and trailing quotes and un-escape other quotes.
+        for token in tokens:
+            if token.startswith('"'):
+                token = token[1:]
+
+            if token.endswith('"'):
+                token = token[:-1]
+
+            token = token.replace('""', '"')
+
+            new_tokens.append(token)
+
+        return new_tokens
+
+    def _parse_filter_expression(self, expression, tokens):
+        # Converts the tokens of a possibly conditional expression into 1 or 2
+        # sub expressions for further parsing.
+        #
+        # Examples:
+        #          ('x', '==', 2000) -> exp1
+        #          ('x', '>',  2000, 'and', 'x', '<', 5000) -> exp1 and exp2
+
+        if len(tokens) == 7:
+            # The number of tokens will be either 3 (for 1 expression)
+            # or 7 (for 2  expressions).
+            conditional = tokens[3]
+
+            if re.match('(and|&&)', conditional):
+                conditional = 0
+            elif re.match('(or|\|\|)', conditional):
+                conditional = 1
+            else:
+                warn("Token '%s' is not a valid conditional "
+                     "in filter expression '%s'" % (conditional, expression))
+
+            expression_1 = self._parse_filter_tokens(expression, tokens[0:3])
+            expression_2 = self._parse_filter_tokens(expression, tokens[4:7])
+
+            return expression_1 + [conditional] + expression_2
+        else:
+            return self._parse_filter_tokens(expression, tokens)
+
+    def _parse_filter_tokens(self, expression, tokens):
+        # Parse the 3 tokens of a filter expression and return the operator
+        # and token. The use of numbers instead of operators is a legacy of
+        # Spreadsheet::WriteExcel.
+        operators = {
+            '==': 2,
+            '=': 2,
+            '=~': 2,
+            'eq': 2,
+
+            '!=': 5,
+            '!~': 5,
+            'ne': 5,
+            '<>': 5,
+
+            '<': 1,
+            '<=': 3,
+            '>': 4,
+            '>=': 6,
+        }
+
+        operator = operators.get(tokens[1], None)
+        token = tokens[2]
+
+        # Special handling of "Top" filter expressions.
+        if re.match('top|bottom', tokens[0].lower()):
+            value = int(tokens[1])
+
+            if (value < 1 or value > 500):
+                warn("The value '%d' in expression '%s' "
+                     "must be in the range 1 to 500" % (value, expression))
+
+            token = token.lower()
+
+            if token != 'items' and token != '%':
+                warn("The type '%s' in expression '%s' "
+                     "must be either 'items' or '%'" % (token, expression))
+
+            if tokens[0].lower() == 'top':
+                operator = 30
+            else:
+                operator = 32
+
+            if tokens[2] == '%':
+                operator += 1
+
+            token = str(value)
+
+        if not operator and tokens[0]:
+            warn("Token '%s' is not a valid operator "
+                 "in filter expression '%s'" % (token[0], expression))
+
+        # Special handling for Blanks/NonBlanks.
+        if re.match('blanks|nonblanks', token.lower()):
+            # Only allow Equals or NotEqual in this context.
+            if operator != 2 and operator != 5:
+                warn("The operator '%s' in expression '%s' "
+                     "is not valid in relation to Blanks/NonBlanks'"
+                     % (tokens[1], expression))
+
+            token = token.lower()
+
+            # The operator should always be 2 (=) to flag a "simple" equality
+            # in the binary record. Therefore we convert <> to =.
+            if token == 'blanks':
+                if operator == 5:
+                    token = ' '
+            else:
+                if operator == 5:
+                    operator = 2
+                    token = 'blanks'
+                else:
+                    operator = 5
+                    token = ' '
+
+        # if the string token contains an Excel match character then change the
+        # operator type to indicate a non "simple" equality.
+        if operator == 2 and re.search('[*?]', token):
+            operator = 22
+
+        return [operator, token]
+
     ###########################################################################
     #
     # XML methods.
@@ -2377,7 +2704,6 @@ class Worksheet(xmlwriter.XMLwriter):
         self._xml_start_tag('hyperlinks')
 
         for args in (hlink_refs):
-            # TODO
             link_type = args.pop(0)
 
             if link_type == 1:
@@ -2420,3 +2746,123 @@ class Worksheet(xmlwriter.XMLwriter):
         attributes.append(('display', display))
 
         self._xml_empty_tag('hyperlink', attributes)
+
+    def _write_auto_filter(self):
+        # Write the <autoFilter> element.
+        if not self.autofilter_ref:
+            return
+
+        attributes = [('ref', self.autofilter_ref)]
+
+        if self.filter_on:
+            # Autofilter defined active filters.
+            self._xml_start_tag('autoFilter', attributes)
+            self._write_autofilters()
+            self._xml_end_tag('autoFilter')
+
+        else:
+            # Autofilter defined without active filters.
+            self._xml_empty_tag('autoFilter', attributes)
+
+    def _write_autofilters(self):
+        # Function to iterate through the columns that form part of an
+        # autofilter range and write the appropriate filters.
+        (col1, col2) = self.filter_range
+
+        for col in range(col1, col2 + 1):
+            # Skip if column doesn't have an active filter.
+            if not col in self.filter_cols:
+                continue
+
+            # Retrieve the filter tokens and write the autofilter records.
+            tokens = self.filter_cols[col]
+            filter_type = self.filter_type[col]
+
+            # Filters are relative to first column in the autofilter.
+            self._write_filter_column(col - col1, filter_type, tokens)
+
+    def _write_filter_column(self, col_id, filter_type, filters):
+        # Write the <filterColumn> element.
+        attributes = [('colId', col_id)]
+
+        self._xml_start_tag('filterColumn', attributes)
+
+        if filter_type == 1:
+            # Type == 1 is the new XLSX style filter.
+            self._write_filters(filters)
+        else:
+            # Type == 0 is the classic "custom" filter.
+            self._write_custom_filters(filters)
+
+        self._xml_end_tag('filterColumn')
+
+    def _write_filters(self, filters):
+        # Write the <filters> element.
+
+        if len(filters) == 1 and filters[0] == 'blanks':
+            # Special case for blank cells only.
+            self._xml_empty_tag('filters', [('blank', 1)])
+        else:
+            # General case.
+            self._xml_start_tag('filters')
+
+            for autofilter in (filters):
+                self._write_filter(autofilter)
+
+            self._xml_end_tag('filters')
+
+    def _write_filter(self, val):
+        # Write the <filter> element.
+        attributes = [('val', val)]
+
+        self._xml_empty_tag('filter', attributes)
+
+    def _write_custom_filters(self, tokens):
+        # Write the <customFilters> element.
+        if len(tokens) == 2:
+            # One filter expression only.
+            self._xml_start_tag('customFilters')
+            self._write_custom_filter(*tokens)
+            self._xml_end_tag('customFilters')
+        else:
+            # Two filter expressions.
+            attributes = []
+
+            # Check if the "join" operand is "and" or "or".
+            if tokens[2] == 0:
+                attributes = [('and', 1)]
+            else:
+                attributes = [('and', 0)]
+
+            # Write the two custom filters.
+            self._xml_start_tag('customFilters', attributes)
+            self._write_custom_filter(tokens[0], tokens[1])
+            self._write_custom_filter(tokens[3], tokens[4])
+            self._xml_end_tag('customFilters')
+
+    def _write_custom_filter(self, operator, val):
+        # Write the <customFilter> element.
+        attributes = []
+
+        operators = {
+            1: 'lessThan',
+            2: 'equal',
+            3: 'lessThanOrEqual',
+            4: 'greaterThan',
+            5: 'notEqual',
+            6: 'greaterThanOrEqual',
+            22: 'equal',
+        }
+
+        # Convert the operator from a number to a descriptive string.
+        if operators[operator] is not None:
+            operator = operators[operator]
+        else:
+            warn("Unknown operator = %s" % operator)
+
+        # The 'equal' operator is the default attribute and isn't stored.
+        if not operator == 'equal':
+            attributes.append(('operator', operator))
+        attributes.append(('val', val))
+
+        self._xml_empty_tag('customFilter', attributes)
